// module playw.cpp
/*-
   All rights reserved.
   This file is distributed under the 2-clause BSD License.
*/
// This is based on  wavplay.c by Zhihao Yuan 
// Modifications and enhancements by E.Huckert 11-2016:
//            sources from zip file wavplay-master.zip (Internet)
//            reformatted
//            all non ALSA compile switches eliminated
//            SUN format eliminated
//            basic logging added
//            signal handling added
// compile under Linux/Ubuntu and Linux/Raspian: 
//          g++ -DUSE_ALSA -DUSE_STL -o playw  playw.cpp 
//                  hu_trace.cpp -lasound -lpthread
// tested under Linux: works!

// To stop this program running in the background:
//   kill -9 processnumber

#include <errno.h>
#include <stdlib.h>
#include <string.h>
#include <sys/param.h>
#include <unistd.h>
#include <signal.h>
#include "wavplay.h"

#include "hu_trace.h"

#define ALSA_PCM_NEW_HW_PARAMS_API
#include <alsa/asoundlib.h>

#define BUF_SIZE  4096
#define eputs(s) (fprintf(stderr, "%s: " s "\n", __func__))

#define WAV_FMT_8  SND_PCM_FORMAT_U8
#define AIF_FMT_8  SND_PCM_FORMAT_S8
#define WAV_FMT_16  SND_PCM_FORMAT_S16_LE
#define AIF_FMT_16  SND_PCM_FORMAT_S16_BE
#define WAV_FMT_24  SND_PCM_FORMAT_S24_3LE
#define SUN_FMT_24  SND_PCM_FORMAT_S24_3BE
#define WAV_FMT_32  SND_PCM_FORMAT_S32_LE
#define AIF_FMT_32  SND_PCM_FORMAT_S32_BE
#define WAV_FMT_A_LAW  SND_PCM_FORMAT_A_LAW
#define WAV_FMT_MU_LAW  SND_PCM_FORMAT_MU_LAW
#define WAV_FMT_IMA_ADPCM  SND_PCM_FORMAT_IMA_ADPCM

static snd_pcm_t *pcm = NULL;
static HU_TRACE *pTrace = NULL;
static struct sigaction act;

// -----------------------------------------
int snd_drop(void) 
{
  return 
  snd_pcm_drop(pcm);
}   // end snd_drop()

// ----------------- communication stuff ------------------------------
void close_everything()
{
  if (::pTrace != NULL)
    ::pTrace->trace("close_everything()");
  snd_drop();
  if (::pTrace != NULL)
  {
     delete ::pTrace;
    ::pTrace = NULL;
  }
}  // end close_everything()

// --------------------------------------
int snd_init(void) 
{
  int ret = 0;
  if (::pTrace != NULL)
    ::pTrace->trace("snd_init() device=", DEV_NAME);
  if (pcm) 
  {
    ret++;
    snd_end();
  }
  ret = snd_pcm_open(&pcm, 
              DEV_NAME,  // "default" in wavplay.h 
              SND_PCM_STREAM_PLAYBACK, 0);
  if (::pTrace != NULL)
    ::pTrace->trace("snd_init() ret=", ret);
}   // end snd_init()

// ---------------------------------------------------
void sigusr_handler(int sig)
{
  if (::pTrace != NULL)
    ::pTrace->trace("sigusr_handler() SIGUSR1 received - we stop");
  exit(0);
}   // end sigusr_handler()

// ---------------------------------------------------
void sigterm_handler(int sig)
{
  if (::pTrace != NULL)
    ::pTrace->trace("sigterm_handler() SIGTERM received - we stop");
  close_everything();
  exit(0);
}   // end sigterm_handler()

// ---------------------------------------------------
void sigquit_handler(int sig)
{
  if (::pTrace != NULL)
    ::pTrace->trace("sigquit_handler() SIGQUIT received - we stop");
  close_everything();
  exit(0);
}   // end sigquit_handler()

// -------------------------------------------------
void setup_signal_handlers()
{
  int    ret;
  //
  if (::pTrace != NULL)
    ::pTrace->trace("setup_signal_handlers()");
  sigfillset(&(act.sa_mask));
  act.sa_handler = sigusr_handler;
  sigaction(SIGUSR1, &act, NULL);  // normally 30 or 10
  //
  //  tested: SIGQUIT is OK
  sigfillset(&(act.sa_mask));
  act.sa_handler = sigquit_handler;
  sigaction(SIGQUIT, &act, NULL);    // normally 3
  //
  // tested: SIGTERM is OK
  sigfillset(&(act.sa_mask));
  act.sa_handler = sigterm_handler;
  sigaction(SIGTERM, &act, NULL);    //  normally 15
  //
}   // end setup_signal_handlers()

// --------------------------------------
int snd_set(int format, int nchannels, int framerate) 
{
  int st = 0;
  unsigned int val = framerate;
  snd_pcm_hw_params_t *params;
  snd_pcm_hw_params_alloca(&params);
  snd_pcm_hw_params_any(pcm, params);
  st |= snd_pcm_hw_params_set_access(pcm, params, SND_PCM_ACCESS_RW_INTERLEAVED);
  st |= snd_pcm_hw_params_set_format(pcm, params, 
                              (snd_pcm_format_t)format);
  st |= snd_pcm_hw_params_set_channels(pcm, params, nchannels);
  st |= snd_pcm_hw_params_set_rate_near(pcm, params, &val, 0);
  st |= snd_pcm_hw_params(pcm, params);
  return st;
}   // end snd_set()

// --------------------------------------
int snd_end(void) 
{
  int st = snd_pcm_close(pcm);
  pcm = NULL;
  return st;
}

// --------------------------------------
// This is the original player program from wavplay.c
int snd_send(FILE *fp, size_t n) 
{
  snd_pcm_format_t format;
  unsigned int nchannels;
  snd_pcm_uframes_t period;
  snd_pcm_hw_params_t *params;
  snd_pcm_hw_params_alloca(&params);
  snd_pcm_hw_params_current(pcm, params);
  snd_pcm_hw_params_get_format(params, &format);
  snd_pcm_hw_params_get_channels(params, &nchannels);
  snd_pcm_hw_params_get_period_size(params, &period, 0);
  int framesize = snd_pcm_format_width(format) / 8 * nchannels;
  unsigned char buf[period * framesize * 128];
  size_t l;
  //
  if (::pTrace != NULL)
    ::pTrace->trace("in snd_send()");
  unsigned count = 0;
  //
  while (n > sizeof(buf)) 
  {
    if ((l = fread(buf, 1, sizeof(buf), fp))) 
    {
      count++;
      //if (::pTrace != NULL)
      //   ::pTrace->trace("   snd_send() ", count, (long)l);
      switch (snd_pcm_writei(pcm, buf, l / framesize)) 
      {
        case -EBADF:
        return -1;
        case -EPIPE:
        #ifndef NDEBUG
          snd_pcm_recover(pcm, -EPIPE, 0);
        #else
          snd_pcm_prepare(pcm);
        #endif
      }
    }
    else goto EOS;
    n -= l;
  }  // end while ..
  if ((l = fread(buf, 1, n, fp)))
  snd_pcm_writei(pcm, buf, l / framesize);
  if (l < n)
    EOS: 
    if (ftell(fp) > 0) 
       eputs("Unexpected end of stream");
  return 
    snd_pcm_drain(pcm);
}   // end snd_send()


// --------------------------------------
static int wav2format(const wavheader_t *wav) 
{
  switch (wav->format) 
  {
    case 1:
    case -2:
      switch ((wav->bitdepth + 7) / 8) 
      {
        case 1: return WAV_FMT_8;
        case 2: return WAV_FMT_16;
  #ifdef  WAV_FMT_24
        case 3: return WAV_FMT_24;
  #endif
  #ifdef  WAV_FMT_32
        case 4: return WAV_FMT_32;
  #endif
        default: return -1;
      }
    case 258:
    case 6: return WAV_FMT_A_LAW;
    case 257:
    case 7: return WAV_FMT_MU_LAW;
    case 17: return wav->bitdepth == 4 ? WAV_FMT_IMA_ADPCM : -1;
    default: return -1;
  }
}

#define chktype(s) (!strncmp(aif->comptype, s, 4))

#undef bswap16
#undef bswap32
#undef bswap64

// --------------------------------------
static __inline uint16_t bswap16(uint16_t x)
{
  return (x << 8 | x >> 8);
}

/* $NetBSD: bswap32.c,v 1.1 1997/10/09 15:42:33 bouyer Exp $ */
// --------------------------------------
static __inline uint32_t bswap32(uint32_t x)
{
  return  ((x << 24) & 0xff000000 ) |
                  ((x <<  8) & 0x00ff0000 ) |
                  ((x >>  8) & 0x0000ff00 ) |
                  ((x >> 24) & 0x000000ff );
}

// --------------------------------------
static __inline uint64_t bswap64(uint64_t x)
{  
  uint32_t *p = (uint32_t *)(void *)&x;
  uint32_t t;
  t = bswap32(p[0]);
  p[0] = bswap32(p[1]);
  p[1] = t;
  return x;
}

/*
 * converts any endianess to host
 * returns -1 when format is wrong,
 *          0 when no convertion is needed,
 *          positive value == len(fmt).
 */
// --------------------------------------
static int endian2h(const char fmt[], void *p) 
{
#if BYTE_ORDER == BIG_ENDIAN
  if (fmt[0] == '>') return 0;
  else if (fmt[0] != '<') return -1;
#else
  if (fmt[0] == '<') return 0;
  else if (fmt[0] != '>') return -1;
#endif
  const char *i = fmt;
  while (*(++i) != '\0') 
  {
#define fmtInt(x, T) \
  case x: { \
          uint##T##_t *n = (uint##T##_t *)p; \
          *n = bswap##T(*n); \
          p += sizeof(uint##T##_t); \
          break; \
  }
#define fmtSkip(x, n) \
  case x: { \
          p += n; \
          break; \
  }
          switch (*i) 
          {
            fmtInt('h', 16);
            fmtInt('l', 32);
            fmtInt('q', 64);
            fmtSkip('.', 1);
            default : return -1;
          }
#undef fmtInt
#undef fmtSkip
  }
  return i - fmt;
}

typedef union 
{
  uint32_t  com[4];
  long double  value;
} ldouble_t;

// --------------------------------------
static __inline long double ext2l(extdouble_t x) 
{
    return (ldouble_t)
#if BYTE_ORDER == BIG_ENDIAN
    {{ 0, x.expon, x.himant, x.lomant }}
#else
    {{ x.lomant, x.himant, x.expon, 0 }}
#endif
      .value;
}

// define 4 macros here
#define skip(n) do { \
  if (! fseek(fp, 0, SEEK_CUR)) { \
    fseek(fp, (n), SEEK_CUR); \
    break; \
  } \
  char c[BUF_SIZE]; \
  size_t i = n; \
  ssize_t l; \
  while (i > 0) { \
    l = fread(c, 1, i > sizeof(c) ? sizeof(c) : i, fp); \
    if (l > 0) i -= l; \
    else break; \
  } \
} while (0)

#define read2(t) (fread(&t, sizeof(t), 1, fp))
#define chkid(s) (!strncmp(ck.id, s, 4))

// --------------------------------------
static size_t wavparse(wavheader_t *wav, FILE *fp) 
{
  riffchunk_t ck;
  if (!read2(ck.id) || !chkid("WAVE"))
          eputs("Not a WAVE file");
  else 
  {
    while (read2(ck)) 
    {
      endian2h("<l", &ck.size);
      if (chkid("data"))
              return ck.size;
      ck.size = (ck.size + 1) / 2 * 2;
      if (chkid("fmt ")) 
      {
        fread(wav, sizeof(wavheader_t), 1, fp);
        if (ck.size < sizeof(wavheader_t))
                eputs("Bad format chunk");
        else
                skip(ck.size - sizeof(wavheader_t));
      }
      else skip(ck.size);
    }
    eputs("Malformed WAVE file");
  }
  return 0;
}

// --------------------------------------
static size_t aifparse(aifheader_t *aif, FILE *fp) 
{
  riffchunk_t ck;
  if (!read2(ck.id) || !(chkid("AIFF") || chkid("AIFC")))
          eputs("Not an AIFF/AIFC file");
  else 
  {
    while (read2(ck)) 
    {
      endian2h(">l", &ck.size);
      if (chkid("SSND"))
              return ck.size;
      ck.size = (ck.size + 1) / 2 * 2;
      if (chkid("COMM")) 
      {
        size_t ckl = MIN(ck.size, sizeof(aifheader_t));
        fread(aif, ckl, 1, fp);
        if (ckl < (sizeof(aifheader_t) - sizeof(aif->comptype)))
                eputs("Bad common chunk");
        else
                skip(ck.size - ckl);
      }
      else skip(ck.size);
    }
    eputs("Malformed AIFF/AIFC file");
  }
  return 0;
}   // end aifparse()

// --------------------------------------
static size_t wav_readinfo(wav_info_t *info, FILE *fp) 
{
  riffchunk_t ck;
  if (read2(ck)) 
  {
    if (chkid("RIFF")) 
    {
      wavheader_t wav[1];
      size_t sz = wavparse(wav, fp);
      endian2h("<hhllhh", wav);
      info->nchannels = wav->nchannels;
      info->framerate = wav->framerate;
      info->sampwidth = (wav->bitdepth + 7) / 8;
      info->nframes = sz / (info->nchannels * info->sampwidth);
      info->devformat = wav2format(wav);
      return sz;
    }
  }
  eputs("Unknown file format");
  return 0;
}   // end wav_readinfo()

#undef skip
#undef read2
#undef chkid

// --------------------------------------
static int wav_setdev(const wav_info_t *info) 
{
  if (::pTrace != NULL)
    ::pTrace->trace("in wav_setdev()");
  if (info->devformat < 0)
    eputs("Unsupported PCM format");
  else if (snd_set(info->devformat, info->nchannels, info->framerate))
    eputs("Failed to setup the sound device");
  else return 0;
  return -1;
}   // end wav_setdev()

// -------------------------------------------
// entry point from file play.c
int wav_send(FILE *fp)    // FILE pointer!
{
  wav_info_t info[1];
  if (::pTrace != NULL)
    ::pTrace->trace("in wav_send()");
  size_t size = wav_readinfo(info, fp);
  if (size && !wav_setdev(info))
     return snd_send(fp, size);
  return -1;
}   // end wav_send()

// ----------------------------------------
// entry point from file play.c
// play a wave file
int wav_play(const char *fn)   // file name!
{
  int st = -1;
  FILE *fp = fopen(fn, "rb");
  unsigned long pid = getpid();
  unsigned long ppid = getppid();  
  if (::pTrace != NULL)
    ("in wav_play() pid,ppid", pid, ppid);
  if (fp)
  {
    st = wav_send(fp);
    fclose(fp);
  } 
  else 
    perror(__func__);
  return st;
}   // end wav_play()

// ----------------------------------------------------------------------
int main(int argc, char *argv[]) 
{
  char *prog = argv[0];
  if (argc < 2) 
  {
    printf("Usage: %s [FILE|-]...\n", prog);
    return 1;
  }
  //
  ::pTrace = new HU_TRACE("playwav1", "localhost", 9514);
  //
  if (snd_init() < 0) 
  {
    fprintf(stderr, "%s: Failed to open device `"
        DEV_NAME "'\n", prog);
    return 1;
  }
  //
  setup_signal_handlers();
  //
  int i = 1;
  int ret = 0;
  //
  while (1) 
  {
    const char *fn = argv[i++];
    if (strcmp(fn, "-") == 0)
      fn = NULL;
    if (fn != NULL)
       ret = wav_play(fn);
    else
    ret = wav_send(stdin);
    //if (fn ? wav_play(fn) : wav_send(stdin))  // EH: reformulated
    if (ret < 0)
      fprintf(stderr, "%s: Skipping file `%s'\n",
          prog, fn ? fn : "STDIN");
    if (i < argc)
      sleep(1);
    else break;
  }
  // note: trace must be deleted last!
  if (::pTrace != NULL)
    delete ::pTrace;
  snd_end();
  return 0;
}   // end main()
